// This file is part of the mcl project.
// Copyright (c) 2022 merryhime
// SPDX-License-Identifier: MIT

#pragma once

namespace mcl::mp {

/// Contains a list of types
template<class... E>
struct list {};

}  // namespace mcl::mp
