/* This file is part of the dynarmic project.
 * Copyright (c) 2018 MerryMage
 * SPDX-License-Identifier: 0BSD
 */

#pragma once

#include <mcl/stdint.hpp>

namespace Dynarmic::Common::Crypto::SM4 {

u8 AccessSubstitutionBox(u8 index);

}  // namespace Dynarmic::Common::Crypto::SM4
